from __future__ import annotations

from typing import Iterator

import numpy as np
import pytest

from holonote.annotate import Annotator, SQLiteDB, UUIDHexStringKey


@pytest.fixture()
def conn_sqlite_uuid(tmp_path) -> Iterator[SQLiteDB]:
    conn = SQLiteDB(filename=str(tmp_path / "test.db"), primary_key=UUIDHexStringKey())
    yield conn
    try:
        conn.cursor.close()
    except Exception:
        pass
    try:
        conn.con.close()
    except Exception:
        pass


@pytest.fixture()
def annotator_range1d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"TIME": np.datetime64},
        fields=["description"],
        region_types=["Range"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_point1d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"TIME": np.datetime64},
        fields=["description"],
        region_types=["Point"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_range2d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"x": float, "y": float},
        fields=["description"],
        region_types=["Range"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_point2d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"x": float, "y": float},
        fields=["description"],
        region_types=["Point"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def multiple_region_annotator(conn_sqlite_uuid) -> Annotator:
    return Annotator(
        {"TIME": np.datetime64},
        fields=["description"],
        region_types=["Point", "Range"],
        connector=conn_sqlite_uuid,
    )


@pytest.fixture()
def multiple_annotators(conn_sqlite_uuid) -> dict[str, Annotator | SQLiteDB]:
    annotator_range1d = Annotator(
        {"TIME": np.datetime64},
        fields=["description"],
        region_types=["Range"],
        connector=conn_sqlite_uuid,
        init=False,
    )
    annotator_range2d = Annotator(
        {"x": float, "y": float},
        fields=["description"],
        region_types=["Range"],
        connector=conn_sqlite_uuid,
    )

    annotator_range1d.load()
    annotator_range2d.load()
    output = {
        "annotation1d": annotator_range1d,
        "annotation2d": annotator_range2d,
        "conn": conn_sqlite_uuid,
    }
    return output


@pytest.fixture()
def multiple_fields_annotator(conn_sqlite_uuid) -> Annotator:
    conn_sqlite_uuid.fields = ["field1", "field2"]
    anno = Annotator(
        {"TIME": np.datetime64},
        fields=["field1", "field2"],
        connector=conn_sqlite_uuid,
    )
    return anno



from __future__ import annotations

import weakref

import numpy as np
import pandas as pd
import param


class AnnotationTable(param.Parameterized):
    """
    Class that stores and manipulates annotation data, including methods
    to declare annotations and commit edits back to the original data
    source such as a database.
    """
    columns = ("region_type", "dim1", "dim2", "value", "_id")

    index = param.List(default=[])

    def __init__(self,  **params):
        """
        Either specify annotation fields with filled field columns
        (via connector or dataframe) or declare the expected
        field columns if starting with no annotation data.
        """
        super().__init__(**params)

        self._region_df = pd.DataFrame(columns=self.columns)
        self._field_df = None

        self._edits = []
        self._index_mapping = {}

        self._update_index()
        self._field_df_snapshot, self._region_df_snapshot = None, None

        self._annotators = weakref.WeakValueDictionary()

    def load(self, connector=None, fields_df=None, primary_key_name=None, fields=None):
        """
        Load the AnnotationTable from a connector or a fields DataFrame.
        """
        if fields is None:
            fields = []

        if [connector, primary_key_name] == [None,None]:
            raise ValueError('Either a connector instance must be supplied or the primary key name supplied')
        if len(fields) < 1:
            raise ValueError('More than one field column is required')
        primary_key_name = primary_key_name if primary_key_name else connector.primary_key.field_name

        if fields_df:
            fields_df = fields_df[fields].copy() # Primary key/index for annotations
            self._field_df = fields_df
        elif connector:
            self.load_annotation_table(connector, fields)
        elif fields_df is None:
            fields_df = pd.DataFrame(columns=[primary_key_name, *fields])
            fields_df = fields_df.set_index(primary_key_name)
            self._field_df = fields_df

        self.clear_edits()
        self._update_index()

    def register_annotator(self, annotator):
        self._annotators[id(annotator)] = annotator


    # FIXME: Multiple region updates
    def update_annotation_region(self, index):
        region = next(iter(self._annotators.values()))._region
        if region == {}:
            print('No new region selected. Skipping')
            return

        value = region['value']
        mask = self._region_df[self._region_df._id == index]
        assert len(mask)==1, 'TODO: Handle multiple region updates for single index'
        self._region_df.at[mask.index.values[0], 'value'] = value
        self._edits.append({'operation':'update', 'id':index,
                            'fields': None,
                            'region_fields':[]})

    @property
    def has_snapshot(self):
        return self._field_df_snapshot is not None

    def revert_to_snapshot(self):
        "Clears outstanding changes and used to implement an basic undo system."
        if self._field_df_snapshot is None:
            raise Exception('Call snapshot method before calling revert_to_snapshot')
        self._field_df = self._field_df_snapshot
        self._region_df = self._region_df_snapshot
        self.clear_edits()


    def snapshot(self):
        "Saves a snapshot. Expected to only be used after a syncing commit"
        self._field_df_snapshot, self._region_df_snapshot = self._snapshot()


    def _snapshot(self):
        return self._field_df.copy(), self._region_df.copy()

    def _update_index(self):
        index_list = list(self._region_df['_id'])

        #for ind in self._deletions:
        #    if ind in index_list:
        #        index_list.remove(ind)
        self.index = list(set(index_list))


    def _expand_commit_by_id(self, id_val, fields=None, region_fields=None):
        kwargs = self._field_df.loc[[id_val]].to_dict('records')[0]
        if fields:
            kwargs = {k:v for k,v in kwargs.items() if k in fields}
        kwargs[self._field_df.index.name] = id_val
        if region_fields == []:
            return kwargs
        region_rows = self._region_df[self._region_df['_id'] == id_val]
        for _, region in region_rows.iterrows():
            # TODO: Handle region_fields. Use id? or uuid? or hash?
            # Or value + kdim + region_type?
            dims =  [el for el in [region['dim1'], region['dim2']] if el is not None]
            for dim in dims:
                if region['region_type'] == 'Range':
                    if len(dims)==1:
                        start, end = region['value']
                    elif dim==dims[0]:
                        start, end, _, _ = region['value']
                    elif dim==dims[1]:
                        _, _, start, end = region['value']
                    kwargs[f'start_{dim}'] = start
                    kwargs[f'end_{dim}'] = end
                elif region['region_type'] == 'Point':
                    point_dim1, point_dim2 = region['value']
                    if dim==dims[0]: # FIXME timestamp mapping
                        kwargs[f'point_{dim}'] = region['value'][0]
                    elif dim==dims[1]:
                        kwargs[f'point_{dim}'] = region['value'][1]
        return kwargs

    def _expand_save_commits(self, ids):
        return {'field_list':[self._expand_commit_by_id(id_val) for id_val in ids]}

    def _create_commits(self):
        "Expands out the commit history into commit operations"
        fields_inds = set(self._field_df.index)
        region_inds = set(self._region_df['_id'].unique())
        unassigned_inds = fields_inds - region_inds
        if unassigned_inds:
            raise ValueError(f'Following annotations have no associated region: {unassigned_inds}')

        commits = []
        for edit in self._edits:
            operation = edit['operation']
            if operation == 'insert':
                # May be missing due to earlier deletion operation - nothing to do
                if edit['id'] not in self._field_df.index:
                    continue
                kwargs = self._expand_commit_by_id(edit['id'])

            elif operation == 'delete':
                kwargs = {'id_val': edit['id']}
            elif operation == 'update':
                if edit['id'] not in self._field_df.index:
                    continue
                kwargs = self._expand_commit_by_id(edit['id'],
                                                   fields=edit['fields'],
                                                   region_fields=edit['region_fields'])
            elif operation == 'save':
                kwargs = self._expand_save_commits(edit['ids'])
            commits.append({'operation':operation, 'kwargs':kwargs})

        return commits

    def commits(self, connector):
        commits = self._create_commits()
        for commit in commits:
            operation = commit['operation']
            kwargs = connector.transforms[operation](commit['kwargs'])
            getattr(connector,connector.operation_mapping[operation])(**kwargs)

        for annotator in self._annotators.values():
            annotator.annotation_table.clear_edits()

        return commits

    def clear_edits(self, edit_type=None):
        "Clear edit state and index mapping"
        self._edits = []
        self._index_mapping = {}

    def add_annotation(self, regions, **fields):
        "Takes a list of regions or the special value 'annotation-regions' to use attached annotators"
        index_value = fields.pop(self._field_df.index.name)
        if regions == 'annotator-regions':
            regions = [a._region for a in self._annotators.values()]
        if all(el=={} for el in regions):
            print('No new region selected. Skipping')
            return
        self._add_annotation_fields(index_value, fields=fields)

        for region in regions:
            if region == {}:
                continue

            region_type, dim1, dim2, value = (region['region_type'], region['dim1'],
                                              region['dim2'], region['value'])
            if region_type not in ['Range', 'Point']:
                raise Exception('TODO: Currently only supporting Range and Point annotations')

            new_region_row = pd.DataFrame([[region_type, dim1, dim2, value, index_value]], columns=self.columns)
            self._region_df = pd.concat((self._region_df, new_region_row), ignore_index=True)

        self._edits.append({'operation':'insert', 'id':index_value})
        self._update_index()

    def refresh_annotators(self, clear=False):
        for annotator in self._annotators.values():
            annotator.refresh(clear=clear)

    def _add_annotation_fields(self, index_value, fields=None):

        index_name_set = set() if self._field_df.index.name is None else set([self._field_df.index.name])
        unknown_kwargs = set(fields.keys()) - set(self._field_df.columns)
        if unknown_kwargs - index_name_set:
            raise KeyError(f'Unknown fields columns: {list(unknown_kwargs)}')

        new_fields = pd.DataFrame([dict(fields, **{self._field_df.index.name:index_value})])
        new_fields = new_fields.set_index(self._field_df.index.name)
        self._field_df =   pd.concat((self._field_df, new_fields))

    def delete_annotation(self, index):
        if index is None:
            raise ValueError('Deletion index cannot be None')
        self._region_df = self._region_df[self._region_df['_id'] != index] # Could match multiple rows
        self._field_df = self._field_df.drop(index, axis=0)

        self._edits.append({'operation':'delete', 'id':index})
        self._update_index()

    def update_annotation_fields(self, index, **fields):
        for column, value in fields.items():
            self._field_df.loc[index][column] = value

        self._edits.append({'operation':'update', 'id':index,
                            'fields' : [c for c in fields.keys()],
                            'region_fields' : []})

    # Methods to map into holoviews

    def _validate_index_to_fields(self, series):
        if series.index.name !=  self._field_df.index.name:
            raise ValueError(f'Index name {series.index.name} does not match fields index name {self._field_df.index.name}')
        if series.index.dtype != self._field_df.index.dtype:
            raise ValueError('Index dtype does not match fields index dtype')

    def _assert_indices_match(self, *series):
        if all(s.index is series[0].index for s in series):
            pass
        else:
            index_names = [s.index.name for s in series]
            if not all(name == index_names[0] for name in index_names):
                raise ValueError(f'Index names do not match: {index_names}')
            # TODO: Match dtypes
            match_values = all(all(s.index == series[0].index) for s in series)
            if not match_values:
                raise ValueError('Indices do not match')

        # for s in series:
        #     self._validate_index_to_fields(s)


    def define_fields(self, fields_df, index_mapping):
        # Need a staging area to hold everything till initialized
        self._index_mapping.update(index_mapping)  # Rename _field_df
        self._field_df = pd.concat([self._field_df, fields_df])
        self._edits.append({'operation':'save', 'ids':list(fields_df.index)})

    def define_points(self, dims, posx, posy=None):
        """
        Points in 1- or 2-dimensional space.

        Both posx and posy expect a Series object with an index
        corresponding to the fields_df supplied in the constructor (if
        it was specified).
        """
        if isinstance(dims, str):
            dims = (dims,)
        if posy is None and len(dims)==2:
            raise ValueError('Two dimensions declared but data for only one specified')
        if posy is not None and len(dims)==1:
            raise ValueError('Only one dimensions declared but data for more than one specified.')

        if len(dims)==2:
            self._assert_indices_match(posx, posy)

        mismatches = [el for el in posx.index if self._index_mapping.get(el,el)
                      not in self._field_df.index]
        if any(mismatches):
            raise KeyError(f'Keys {mismatches} do not match any fields entries')

        dim2 = None if len(dims)==1 else dims[1]
        value = zip(posx,  pd.Series([None for el in range(len(posx))])) if len(dims)==1 else zip(posx, posy)
        additions = pd.DataFrame({"region_type":'Point',
                                  "dim1":dims[0],
                                  "dim2":dim2,
                                  "value":value,
                                  "_id":[self._index_mapping[ind] for ind in posx.index]})
        self._region_df = pd.concat((self._region_df, additions), ignore_index=True)

    def define_ranges(self, dims, startx, endx, starty=None, endy=None):
        if isinstance(dims, str):
            dims = (dims,)
        if len(dims)==2 and any([el is None for el in [starty, endy]]):
            raise ValueError('Two dimensions declared but insufficient data specified')
        if len(dims)==1 and (starty, endy) != (None, None):
            raise ValueError('Only one dimensions declared but data for more than one specified.')

        if len(dims)==1:
            self._assert_indices_match(startx, endx)
        else:
            self._assert_indices_match(startx, endx, starty, endy)

        mismatches = [el for el in startx.index if self._index_mapping.get(el,el)
                      not in self._field_df.index]
        if any(mismatches):
            raise KeyError(f'Keys {mismatches} do not match any fields entries')


        dim2 = None if len(dims)==1 else dims[1]
        value = zip(startx, endx) if len(dims)==1 else zip(startx, endx, starty, endy)
        additions = pd.DataFrame({"region_type":'Range',
                                  "dim1":dims[0],
                                  "dim2":dim2,
                                  "value":value,
                                  "_id":[self._index_mapping[ind] for ind in startx.index]})
        self._region_df = pd.concat((self._region_df, additions), ignore_index=True)
        self._update_index()


    def _filter(self, dim_mask, region_type):
        region_mask = self._region_df["region_type"] == region_type
        return self._region_df[np.logical_and(region_mask, dim_mask)]

    def _mask1D(self, kdims):
        return np.logical_and(
            self._region_df["dim1"] == str(kdims[0]), self._region_df["dim2"].isnull()
        )

    def _mask2D(self, kdims):
        dim1_name, dim2_name = str(kdims[0]), str(kdims[1])
        return np.logical_and(
            self._region_df["dim1"] == dim1_name, self._region_df["dim2"] == dim2_name
        )

    def load_annotation_table(self, conn, fields):
        df = conn.transforms['load'](conn.load_dataframe())
        fields_df = df[fields].copy()
        self.define_fields(fields_df, {ind:ind for ind in fields_df.index})
        all_region_types = [an.region_types for an in self._annotators.values()]
        all_kdim_dtypes = [an.kdim_dtypes for an in self._annotators.values()]
        for region_types, kdim_dtypes in zip(all_region_types, all_kdim_dtypes):
            assert all(el in ['Range', 'Point'] for el in region_types)
            for region_type in region_types:
                if len(kdim_dtypes)==1:
                    kdim = next(iter(kdim_dtypes.keys()))
                    if region_type == 'Range':
                        expected_keys = [f'start_{kdim}', f'end_{kdim}']
                        conn._incompatible_schema_check(expected_keys, list(df.columns), fields, region_type)
                        self.define_ranges(kdim, df[f'start_{kdim}'], df[f'end_{kdim}'])
                    elif region_type == 'Point':
                        conn._incompatible_schema_check([f'point_{kdim}'], list(df.columns), fields, region_type)
                        self.define_points(kdim, df[f'point_{kdim}'])
                elif len(kdim_dtypes)==2:
                    kdim1, kdim2 = list(kdim_dtypes.keys())
                    if region_type == 'Range':
                        conn._incompatible_schema_check([f'start_{kdim1}', f'end_{kdim1}',
                                                         f'start_{kdim2}', f'end_{kdim2}'],
                                                        list(df.columns), fields, region_type)
                        self.define_ranges([kdim1, kdim2],
                                                       df[f'start_{kdim1}'], df[f'end_{kdim1}'],
                                                       df[f'start_{kdim2}'], df[f'end_{kdim2}'])
                    elif region_type == 'Point':
                        conn._incompatible_schema_check([f'point_{kdim1}', f'point_{kdim2}'],
                                                        list(df.columns), fields, region_type)
                        self.define_points([kdim1, kdim2], df[f'point_{kdim1}'], df[f'point_{kdim2}'])
        self.clear_edits()

    def add_schema_to_conn(self, conn):
        field_dtypes = {col: str for col in conn.fields} # FIXME - generalize
        all_region_types = [an.region_types for an in self._annotators.values()]
        all_kdim_dtypes = [an.kdim_dtypes for an in self._annotators.values()]
        schema = conn.generate_schema(conn.primary_key, all_region_types, all_kdim_dtypes, field_dtypes)
        conn.add_schema(schema)



from __future__ import annotations

from .annotator import Annotator  # noqa: F401
from .connector import (  # noqa: F401
    AutoIncrementKey,
    Connector,
    SQLiteDB,
    UUIDBinaryKey,
    UUIDHexStringKey,
)
from .table import *



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from holonote.annotate import (
    AutoIncrementKey,
    Connector,
    UUIDBinaryKey,
    UUIDHexStringKey,
)


@pytest.fixture(params=[UUIDHexStringKey, AutoIncrementKey, UUIDBinaryKey])
def database(conn_sqlite_uuid, request):
    # Change the primary key type
    conn_sqlite_uuid.primary_key = request.param(field_name='uuid')
    fields = {
        'uuid': conn_sqlite_uuid.primary_key.schema,
        'description': 'TEXT',
        'start':'TIMESTAMP',
        'end': 'TIMESTAMP'
    }
    conn_sqlite_uuid.initialize(fields)
    return conn_sqlite_uuid


class TestConnector:
    "Tests for classmethods on the base class"

    def test_fields_from_metadata_literals(self):
        fields = Connector.schema_from_field_values({'A':3, 'B':'string', 'C':False})
        assert fields == {'A': 'INTEGER', 'B': 'TEXT', 'C': 'BOOLEAN'}

    def test_schema_from_value_datetime(self):
        datetime_type = Connector.field_value_to_type(np.datetime64('NaT'))
        assert Connector.type_mapping[datetime_type] == 'TIMESTAMP'

    def test_expand_range_region_column_schema_datetime(self):
        result = Connector.expand_region_column_schema(['Range'], {'xdim':np.datetime64})
        assert result == {'start_xdim': 'TIMESTAMP', 'end_xdim': 'TIMESTAMP'}

    def test_expand_range_region_column_schema_datetimes(self):
        result = Connector.expand_region_column_schema(['Range'],
                                                       {'xdim':np.datetime64,
                                                        'ydim':int})
        expected = {
            'start_xdim': 'TIMESTAMP', 'end_xdim': 'TIMESTAMP',
            'start_ydim': 'INTEGER', 'end_ydim': 'INTEGER'
        }
        assert result == expected

    def test_generate_schema(self):
        region_types = [['Range']]
        kdim_dtypes = [{'xdim':np.datetime64, 'ydim':int}]
        result = Connector.generate_schema(AutoIncrementKey(), region_types, kdim_dtypes,
                                           {'description':str})
        expected = {
            'id': 'INTEGER PRIMARY KEY AUTOINCREMENT', 'start_xdim': 'TIMESTAMP',
            'start_ydim': 'INTEGER', 'end_xdim': 'TIMESTAMP',
            'end_ydim': 'INTEGER', 'description': 'TEXT'
        }
        assert result == expected


class TestSQLiteDB:

    def test_setup(self, database):
        assert database.con is not None

    def test_initialized(self, database):
        assert not database.uninitialized

    def test_add_row(self, database, request):
        id1 = database.primary_key(database)
        start = pd.Timestamp('2022-06-01')
        end = pd.Timestamp('2022-06-03')
        description = 'A description'
        insertion = {"uuid": id1, 'description':description, 'start':start, 'end':end}
        df = pd.DataFrame({"uuid":pd.Series([id1], dtype=object),
                           'description':[description], 'start':[start], 'end':[end]}).set_index("uuid")
        database.add_row(**insertion)
        pd.testing.assert_frame_equal(database.load_dataframe(), df)

    def test_add_three_rows_delete_one(self, database):
        id1 = database.primary_key(database)
        insertion1 = {'uuid': id1,
                     'description':'A description',
                     'start':pd.Timestamp('2022-06-01'),
                     'end':pd.Timestamp('2022-06-03')}

        id2 = database.primary_key(database, [id1])
        insertion2 = {'uuid': id2,
                     'description':'A 2nd description',
                     'start':pd.Timestamp('2024-06-01'),
                     'end':pd.Timestamp('2024-06-03')}

        id3 = database.primary_key(database, [id2])
        insertion3 = {'uuid': id3,
                     'description':'A 3rd description',
                     'start':pd.Timestamp('2026-06-01'),
                     'end':pd.Timestamp('2026-06-03')}

        df_data = {'uuid': pd.Series([insertion1['uuid'], insertion3['uuid']], dtype=object),
                   'description':[insertion1['description'], insertion3['description']],
                   'start':[insertion1['start'], insertion3['start']],
                   'end':[insertion1['end'], insertion3['end']]}
        df = pd.DataFrame(df_data).set_index('uuid')
        database.add_row(**insertion1)
        database.add_row(**insertion2)
        database.add_row(**insertion3)
        database.delete_row(id2)
        pd.testing.assert_frame_equal(database.load_dataframe(), df)



from __future__ import annotations

import uuid

import numpy as np
import pandas as pd
import pytest

# TODO:

# * (after refactor) annotators -> annotator, connectors -> connector [ ]
# TESTS
# Schema error (needs file or connect in memory??)
# .snapshot() and .revert_to_snapshot()


class TestBasicRange1DAnnotator:
    def test_point_insertion_exception(self, annotator_range1d):
        timestamp = np.datetime64('2022-06-06')
        expected_msg = r"Point region types not enabled as region_types=\['Range'\]"
        with pytest.raises(ValueError, match=expected_msg):
            annotator_range1d.set_point(timestamp)

    def test_insertion_edit_table_columns(self, annotator_range1d):
        annotator_range1d.set_range(np.datetime64('2022-06-06'), np.datetime64('2022-06-08'))
        annotator_range1d.add_annotation(description='A test annotation!')
        commits = annotator_range1d.commit(return_commits=True)
        assert len(commits)==1, 'Only one insertion commit made '
        annotator_range1d.commit(return_commits=True)
        assert commits[0]['operation'] == 'insert'
        assert set(commits[0]['kwargs'].keys()) == set(annotator_range1d.connector.columns)

    def test_range_insertion_values(self, annotator_range1d) -> None:
        start, end = np.datetime64('2022-06-06'), np.datetime64('2022-06-08')
        annotator_range1d.set_range(start, end)
        annotator_range1d.add_annotation(description='A test annotation!')
        commits = annotator_range1d.commit(return_commits=True)
        assert len(commits)==1, 'Only one insertion commit made'
        kwargs = commits[0]['kwargs']
        assert 'uuid' in kwargs.keys(), 'Expected uuid primary key in kwargs'
        kwargs.pop('uuid')
        assert kwargs, dict(description='A test annotation!', start_TIME=start, end_TIME=end)

    def test_range_commit_insertion(self, annotator_range1d):
        start, end  = np.datetime64('2022-06-06'), np.datetime64('2022-06-08')
        description = 'A test annotation!'
        annotator_range1d.set_range(start, end)
        annotator_range1d.add_annotation(description=description)
        annotator_range1d.commit(return_commits=True)

        df = pd.DataFrame({'uuid': pd.Series(annotator_range1d.df.index[0], dtype=object),
                           'start_TIME':[start],
                           'end_TIME':[end],
                           'description':[description]}
                           ).set_index('uuid')

        sql_df = annotator_range1d.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)


    def test_range_addition_deletion_by_uuid(self, annotator_range1d):
        start1, end1  = np.datetime64('2022-06-06'), np.datetime64('2022-06-08')
        start2, end2  = np.datetime64('2023-06-06'), np.datetime64('2023-06-08')
        start3, end3  = np.datetime64('2024-06-06'), np.datetime64('2024-06-08')
        annotator_range1d.set_range(start1, end1)
        annotator_range1d.add_annotation(description='Annotation 1')
        annotator_range1d.set_range(start2, end2)
        annotator_range1d.add_annotation(description='Annotation 2', uuid='08286429')
        annotator_range1d.set_range(start3, end3)
        annotator_range1d.add_annotation(description='Annotation 3')
        annotator_range1d.commit(return_commits=True)
        sql_df = annotator_range1d.connector.load_dataframe()
        assert set(sql_df['description']) ==set(['Annotation 1', 'Annotation 2', 'Annotation 3'])
        deletion_index = sql_df.index[1]
        annotator_range1d.delete_annotation(deletion_index)
        annotator_range1d.commit(return_commits=True)
        sql_df = annotator_range1d.connector.load_dataframe()
        assert set(sql_df['description']) == set(['Annotation 1', 'Annotation 3'])


    def test_range_define_preserved_index_mismatch(self, annotator_range1d):
        starts = [np.datetime64('2022-06-%.2d' % d) for d in  range(6,15, 4)]
        ends = [np.datetime64('2022-06-%.2d' % (d+2)) for d in  range(6,15, 4)]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id = [uuid.uuid4().hex[:8] for d in [1,2,3]]

        data = pd.DataFrame({'uuid':annotation_id, 'start':starts, 'end':ends, 'description':descriptions}).set_index('uuid')
        annotator_range1d.define_fields(data[['description']], preserve_index=True)
        annotator_range1d.define_ranges(data['start'].iloc[:2], data['end'].iloc[:2])
        msg = f"Following annotations have no associated region: {{{annotation_id[2]!r}}}"
        with pytest.raises(ValueError, match=msg):
            annotator_range1d.commit(return_commits=True)

    def test_range_define_auto_index_mismatch(self, annotator_range1d):
        starts = [np.datetime64('2022-06-%.2d' % d) for d in  range(6,15, 4)]
        ends = [np.datetime64('2022-06-%.2d' % (d+2)) for d in  range(6,15, 4)]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id = [uuid.uuid4().hex[:8] for d in [1,2,3]]

        data = pd.DataFrame({'uuid':annotation_id, 'start':starts,
                             'end':ends, 'description':descriptions}).set_index('uuid')
        annotator_range1d.define_fields(data[['description']], preserve_index=False)
        annotator_range1d.define_ranges(data['start'].iloc[:2], data['end'].iloc[:2])
        with pytest.raises(ValueError,
                           match="Following annotations have no associated region:"):
            annotator_range1d.commit(return_commits=True)

    def test_range_define_unassigned_indices(self, annotator_range1d):
        starts = [np.datetime64('2022-06-%.2d' % d) for d in  range(6,15, 4)]
        ends = [np.datetime64('2022-06-%.2d' % (d+2)) for d in  range(6,15, 4)]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id1 = [uuid.uuid4().hex[:8] for d in [1,2,3]]
        mismatched = [uuid.uuid4().hex[:8] for d in [1,2]]
        annotation_id2 = [*mismatched, annotation_id1[2]]

        data1 = pd.DataFrame({'uuid':annotation_id1, 'start':starts,
                              'end':ends, 'description':descriptions}).set_index('uuid')
        data2 = pd.DataFrame({'uuid':annotation_id2, 'start':starts,
                              'end':ends, 'description':descriptions}).set_index('uuid')

        annotator_range1d.define_fields(data1[['description']])
        with pytest.raises(KeyError, match=str(mismatched)):
            annotator_range1d.define_ranges(data2['start'], data2['end'])


class TestBasicRange2DAnnotator:

    def test_point_insertion_exception(self, annotator_range2d):
        x,y = 0.5,0.5
        expected_msg = r"Point region types not enabled as region_types=\['Range'\]"
        with pytest.raises(ValueError, match=expected_msg):
            annotator_range2d.set_point(x,y)

    def test_insertion_edit_table_columns(self, annotator_range2d):
        annotator_range2d.set_range(-0.25, 0.25, -0.1, 0.1)
        annotator_range2d.add_annotation(description='A test annotation!')
        commits = annotator_range2d.commit(return_commits=True)
        assert len(commits)==1, 'Only one insertion commit made '
        assert commits[0]['operation'] == 'insert'
        assert set(commits[0]['kwargs'].keys()) == set(annotator_range2d.connector.columns)

    def test_range_insertion_values(self, annotator_range2d):
        startx, endx, starty, endy = -0.25, 0.25, -0.1, 0.1
        annotator_range2d.set_range(startx, endx, starty, endy)
        annotator_range2d.add_annotation(description='A test annotation!')
        commits = annotator_range2d.commit(return_commits=True)
        assert len(commits)==1, 'Only one insertion commit made'
        kwargs = commits[0]['kwargs']
        assert 'uuid' in kwargs.keys(), 'Expected uuid primary key in kwargs'
        kwargs.pop('uuid')
        assert kwargs ==  dict(description='A test annotation!',
                                      start_x=startx, end_x=endx, start_y=starty, end_y=endy)

    def test_range_commit_insertion(self, annotator_range2d):
        startx, endx, starty, endy = -0.25, 0.25, -0.1, 0.1
        description = 'A test annotation!'
        annotator_range2d.set_range(startx, endx, starty, endy)
        annotator_range2d.add_annotation(description=description)
        annotator_range2d.commit(return_commits=True)

        df = pd.DataFrame({'uuid': pd.Series(annotator_range2d.df.index[0], dtype=object),
                           'start_x':[startx],
                           'start_y':[starty],
                           'end_x':[endx],
                           'end_y':[endy],
                           'description':[description]}
                           ).set_index('uuid')

        sql_df = annotator_range2d.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)


    def test_range_addition_deletion_by_uuid(self, annotator_range2d):
        startx1, endx1, starty1, endy1 = -0.251, 0.251, -0.11, 0.11
        startx2, endx2, starty2, endy2 = -0.252, 0.252, -0.12, 0.12
        startx3, endx3, starty3, endy3 = -0.253, 0.253, -0.13, 0.13
        annotator_range2d.set_range(startx1, endx1, starty1, endy1)
        annotator_range2d.add_annotation(description='Annotation 1')
        annotator_range2d.set_range(startx2, endx2, starty2, endy2)
        annotator_range2d.add_annotation(description='Annotation 2', uuid='08286429')
        annotator_range2d.set_range(startx3, endx3, starty3, endy3)
        annotator_range2d.add_annotation(description='Annotation 3')
        annotator_range2d.commit(return_commits=True)
        sql_df = annotator_range2d.connector.load_dataframe()
        assert set(sql_df['description']) == set(['Annotation 1', 'Annotation 2', 'Annotation 3'])
        deletion_index = sql_df.index[1]
        annotator_range2d.delete_annotation(deletion_index)
        annotator_range2d.commit(return_commits=True)
        sql_df = annotator_range2d.connector.load_dataframe()
        assert set(sql_df['description']) == set(['Annotation 1', 'Annotation 3'])


    def test_range_define_preserved_index_mismatch(self, annotator_range2d):
        xstarts, xends = [-0.3, -0.2, -0.1], [0.3, 0.2, 0.1]
        ystarts, yends = [-0.35, -0.25, -0.15], [0.35, 0.25, 0.15]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id = [uuid.uuid4().hex[:8] for d in [1,2,3]]

        data = pd.DataFrame({'uuid':annotation_id, 'xstart':xstarts, 'xend':xends,
                             'ystart':ystarts, 'yend':yends,
                             'description':descriptions}).set_index('uuid')
        annotator_range2d.define_fields(data[['description']], preserve_index=True)
        annotator_range2d.define_ranges(data['xstart'].iloc[:2], data['xend'].iloc[:2],
                                     data['ystart'].iloc[:2], data['yend'].iloc[:2])

        msg = f"Following annotations have no associated region: {{{annotation_id[2]!r}}}"
        with pytest.raises(ValueError, match=msg):
            annotator_range2d.commit(return_commits=True)

    def test_range_define_auto_index_mismatch(self, annotator_range2d):
        xstarts, xends = [-0.3, -0.2, -0.1], [0.3, 0.2, 0.1]
        ystarts, yends = [-0.35, -0.25, -0.15], [0.35, 0.25, 0.15]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id = [uuid.uuid4().hex[:8] for d in [1,2,3]]
        data = pd.DataFrame({'uuid':annotation_id, 'xstart':xstarts, 'xend':xends,
                             'ystart':ystarts, 'yend':yends,
                             'description':descriptions}).set_index('uuid')
        annotator_range2d.define_fields(data[['description']], preserve_index=False)
        annotator_range2d.define_ranges(data['xstart'].iloc[:2], data['xend'].iloc[:2],
                                     data['ystart'].iloc[:2], data['yend'].iloc[:2])
        msg = "Following annotations have no associated region:"
        with pytest.raises(ValueError, match=msg):
            annotator_range2d.commit(return_commits=True)

    def test_range_define_unassigned_indices(self, annotator_range2d):
        xstarts, xends = [-0.3, -0.2, -0.1], [0.3, 0.2, 0.1]
        ystarts, yends = [-0.35, -0.25, -0.15], [0.35, 0.25, 0.15]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id1 = [uuid.uuid4().hex[:8] for d in [1,2,3]]
        mismatched = [uuid.uuid4().hex[:8] for d in [1,2]]
        annotation_id2 = [*mismatched, annotation_id1[2]]

        data1 = pd.DataFrame({'uuid':annotation_id1, 'xstart':xstarts, 'xend':xends,
                             'ystart':ystarts, 'yend':yends,
                             'description':descriptions}).set_index('uuid')
        data2 = pd.DataFrame({'uuid':annotation_id2, 'xstart':xstarts, 'xend':xends,
                             'ystart':ystarts, 'yend':yends,
                             'description':descriptions}).set_index('uuid')

        annotator_range2d.define_fields(data1[['description']])
        with pytest.raises(KeyError, match=str(mismatched)):
            annotator_range2d.define_ranges(data2['xstart'], data2['xend'],
                                         data2['ystart'], data2['yend'])


class TestBasicPoint1DAnnotator:

    def test_insertion_edit_table_columns(self, annotator_point1d):
        annotator_point1d.set_point(np.datetime64('2022-06-06'))
        annotator_point1d.add_annotation(description='A test annotation!')
        commits = annotator_point1d.commit(return_commits=True)
        assert len(commits)==1, 'Only one insertion commit made '
        annotator_point1d.commit(return_commits=True)
        assert commits[0]['operation'] == 'insert'
        assert set(commits[0]['kwargs'].keys()) == set(annotator_point1d.connector.columns)

    def test_range_insertion_exception(self, annotator_point1d):
        start, end = np.datetime64('2022-06-06'), np.datetime64('2022-06-08')
        msg = r"Range region types not enabled as region_types=\['Point'\]"
        with pytest.raises(ValueError, match=msg):
            annotator_point1d.set_range(start, end)

    def test_point_insertion_values(self, annotator_point1d):
        timestamp = np.datetime64('2022-06-06')
        annotator_point1d.set_point(timestamp)
        annotator_point1d.add_annotation(description='A test annotation!')
        commits = annotator_point1d.commit(return_commits=True)
        assert len(commits)==1, 'Only one insertion commit made'
        kwargs = commits[0]['kwargs']
        assert 'uuid' in kwargs.keys(), 'Expected uuid primary key in kwargs'
        kwargs.pop('uuid')
        assert kwargs == dict(description='A test annotation!', point_TIME=timestamp)

    def test_point_commit_insertion(self, annotator_point1d):
        timestamp = np.datetime64('2022-06-06')
        description = 'A test annotation!'
        annotator_point1d.set_point(timestamp)
        annotator_point1d.add_annotation(description=description)
        annotator_point1d.commit(return_commits=True)

        df = pd.DataFrame({'uuid': pd.Series(annotator_point1d.df.index[0], dtype=object),
                           'point_TIME':[timestamp],
                           'description':[description]}
                           ).set_index('uuid')

        sql_df = annotator_point1d.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)


    def test_point_addition_deletion_by_uuid(self, annotator_point1d):
        ts1  = np.datetime64('2022-06-06')
        ts2  = np.datetime64('2023-06-06')
        ts3  = np.datetime64('2024-06-06')
        annotator_point1d.set_point(ts1)
        annotator_point1d.add_annotation(description='Annotation 1')
        annotator_point1d.set_point(ts2)
        annotator_point1d.add_annotation(description='Annotation 2', uuid='08286429')
        annotator_point1d.set_point(ts3)
        annotator_point1d.add_annotation(description='Annotation 3')
        annotator_point1d.commit(return_commits=True)
        sql_df = annotator_point1d.connector.load_dataframe()
        assert set(sql_df['description']) == set(['Annotation 1', 'Annotation 2', 'Annotation 3'])
        deletion_index = sql_df.index[1]
        annotator_point1d.delete_annotation(deletion_index)
        annotator_point1d.commit(return_commits=True)
        sql_df = annotator_point1d.connector.load_dataframe()
        assert set(sql_df['description']) == set(['Annotation 1', 'Annotation 3'])

    def test_point_define_preserved_index_mismatch(self, annotator_point1d):
        timestamps = [np.datetime64('2022-06-%.2d' % d) for d in  range(6,15, 4)]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id = [uuid.uuid4().hex[:8] for d in [1,2,3]]

        data = pd.DataFrame({'uuid':annotation_id, 'timestamps':timestamps,
                             'description':descriptions}).set_index('uuid')
        annotator_point1d.define_fields(data[['description']], preserve_index=True)
        annotator_point1d.define_points(data['timestamps'].iloc[:2])
        msg = f"Following annotations have no associated region: {{{annotation_id[2]!r}}}"
        with pytest.raises(ValueError, match=msg):
            annotator_point1d.commit(return_commits=True)

    def test_point_define_auto_index_mismatch(self, annotator_point1d):
        timestamps = [np.datetime64('2022-06-%.2d' % d) for d in  range(6,15, 4)]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id = [uuid.uuid4().hex[:8] for d in [1,2,3]]

        data = pd.DataFrame({'uuid':annotation_id, 'timestamps':timestamps,
                             'description':descriptions}).set_index('uuid')
        annotator_point1d.define_fields(data[['description']], preserve_index=False)
        annotator_point1d.define_points(data['timestamps'].iloc[:2])
        with pytest.raises(ValueError, match="Following annotations have no associated region:"):
            annotator_point1d.commit(return_commits=True)

    def test_point_define_unassigned_indices(self, annotator_point1d):
        timestamps = [np.datetime64('2022-06-%.2d' % d) for d in  range(6,15, 4)]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id1 = [uuid.uuid4().hex[:8] for d in [1,2,3]]
        mismatched = [uuid.uuid4().hex[:8] for d in [1,2]]
        annotation_id2 = [*mismatched, annotation_id1[2]]

        data1 = pd.DataFrame({'uuid':annotation_id1, 'timestamps':timestamps,
                              'description':descriptions}).set_index('uuid')
        data2 = pd.DataFrame({'uuid':annotation_id2, 'timestamps':timestamps,
                              'description':descriptions}).set_index('uuid')

        annotator_point1d.define_fields(data1[['description']])
        with pytest.raises(KeyError, match=str(mismatched)):
            annotator_point1d.define_points(data2['timestamps'])


class TestBasicPoint2DAnnotator:

    def test_insertion_edit_table_columns(self, annotator_point2d):
        annotator_point2d.set_point(-0.25, 0.1)
        annotator_point2d.add_annotation(description='A test annotation!')
        commits = annotator_point2d.commit(return_commits=True)
        assert len(commits)==1, 'Only one insertion commit made '
        assert commits[0]['operation'] == 'insert'
        assert set(commits[0]['kwargs'].keys()) == set(annotator_point2d.connector.columns)

    def test_range_insertion_exception(self, annotator_point2d):
        x1,x2,y1,y2 = -0.25,0.25, -0.3, 0.3
        expected_msg = r"Range region types not enabled as region_types=\['Point'\]"
        with pytest.raises(ValueError, match=expected_msg):
            annotator_point2d.set_range(x1,x2,y1,y2)

    def test_point_insertion_values(self, annotator_point2d):
        x,y = 0.5, 0.3
        annotator_point2d.set_point(x,y)
        annotator_point2d.add_annotation(description='A test annotation!')
        commits = annotator_point2d.commit(return_commits=True)
        assert len(commits)==1, 'Only one insertion commit made'
        kwargs = commits[0]['kwargs']
        assert 'uuid' in kwargs.keys(), 'Expected uuid primary key in kwargs'
        kwargs.pop('uuid')
        assert kwargs == dict(description='A test annotation!', point_x=x, point_y=y)

    def test_point_commit_insertion(self, annotator_point2d):
        x, y = 0.5, 0.3
        description = 'A test annotation!'
        annotator_point2d.set_point(x,y)
        annotator_point2d.add_annotation(description=description)
        annotator_point2d.commit(return_commits=True)

        df = pd.DataFrame({'uuid': pd.Series(annotator_point2d.df.index[0], dtype=object),
                           'point_x':[x],
                           'point_y':[y],
                           'description':[description]}
                           ).set_index('uuid')

        sql_df = annotator_point2d.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)


    def test_point_addition_deletion_by_uuid(self, annotator_point2d):
        x1, y1  = 0.2,0.2
        x2, y2  = 0.3,0.3
        x3, y3  = 0.4,0.4
        annotator_point2d.set_point(x1, y1)
        annotator_point2d.add_annotation(description='Annotation 1')
        annotator_point2d.set_point(x2, y2)
        annotator_point2d.add_annotation(description='Annotation 2', uuid='08286429')
        annotator_point2d.set_point(x3, y3)
        annotator_point2d.add_annotation(description='Annotation 3')
        annotator_point2d.commit(return_commits=True)
        sql_df = annotator_point2d.connector.load_dataframe()
        assert set(sql_df['description']) == set(['Annotation 1', 'Annotation 2', 'Annotation 3'])
        deletion_index = sql_df.index[1]
        annotator_point2d.delete_annotation(deletion_index)
        annotator_point2d.commit(return_commits=True)
        sql_df = annotator_point2d.connector.load_dataframe()
        assert set(sql_df['description']) == set(['Annotation 1', 'Annotation 3'])

    def test_point_define_preserved_index_mismatch(self, annotator_point2d):
        xs, ys  = [-0.1,-0.2,-0.3], [0.1,0.2,0.3]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id = [uuid.uuid4().hex[:8] for d in [1,2,3]]

        data = pd.DataFrame({'uuid':annotation_id, 'xs':xs, 'ys':ys,
                             'description':descriptions}).set_index('uuid')
        annotator_point2d.define_fields(data[['description']], preserve_index=True)
        annotator_point2d.define_points(data['xs'].iloc[:2], data['ys'].iloc[:2])
        msg = f"Following annotations have no associated region: {{{annotation_id[2]!r}}}"
        with pytest.raises(ValueError, match=msg):
            annotator_point2d.commit(return_commits=True)

    def test_point_define_auto_index_mismatch(self, annotator_point2d):
        xs, ys  = [-0.1,-0.2,-0.3], [0.1,0.2,0.3]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id = [uuid.uuid4().hex[:8] for d in [1,2,3]]

        data = pd.DataFrame({'uuid':annotation_id, 'xs':xs, 'ys':ys,
                             'description':descriptions}).set_index('uuid')
        annotator_point2d.define_fields(data[['description']], preserve_index=False)
        annotator_point2d.define_points(data['xs'].iloc[:2], data['ys'].iloc[:2])
        msg = "Following annotations have no associated region:"
        with pytest.raises(ValueError, match=msg):
            annotator_point2d.commit(return_commits=True)

    def test_point_define_unassigned_indices(self, annotator_point2d):
        xs, ys  = [-0.1,-0.2,-0.3], [0.1,0.2,0.3]
        descriptions = ['Annotation %d' % d for d in [1,2,3]]
        annotation_id1 = [uuid.uuid4().hex[:8] for d in [1,2,3]]
        mismatched = [uuid.uuid4().hex[:8] for d in [1,2]]
        annotation_id2 = [*mismatched, annotation_id1[2]]

        data1 = pd.DataFrame({'uuid':annotation_id1, 'xs':xs, 'ys':ys,
                              'description':descriptions}).set_index('uuid')
        data2 = pd.DataFrame({'uuid':annotation_id2, 'xs':xs, 'ys':ys,
                              'description':descriptions}).set_index('uuid')

        annotator_point2d.define_fields(data1[['description']])
        with pytest.raises(KeyError, match=str(mismatched)):
            annotator_point2d.define_points(data2['xs'], data2['ys'])



from __future__ import annotations

import holoviews as hv
import numpy as np
import pandas as pd
import pytest

from holonote.annotate import Annotator, SQLiteDB


def test_multipoint_range_commit_insertion(multiple_region_annotator):
    descriptions = ['A point insertion', 'A range insertion']
    timestamp = np.datetime64('2022-06-06')
    multiple_region_annotator.set_point(timestamp)
    multiple_region_annotator.add_annotation(description=descriptions[0])

    start, end = np.datetime64('2022-06-06'), np.datetime64('2022-06-08')
    multiple_region_annotator.set_range(start, end)
    multiple_region_annotator.add_annotation(description=descriptions[1])

    multiple_region_annotator.commit(return_commits=True)

    # FIXME! Index order is inverted?
    df = pd.DataFrame({'uuid': pd.Series(multiple_region_annotator.df.index[::-1], dtype=object),
                        'point_TIME':[timestamp, pd.NaT],
                        'start_TIME':[pd.NaT, start],
                        'end_TIME': [pd.NaT, end],
                        'description':descriptions}
                        ).set_index('uuid')

    sql_df = multiple_region_annotator.connector.load_dataframe()
    pd.testing.assert_frame_equal(sql_df, df)


def test_infer_kdim_dtype_image():
    xvals, yvals  = np.linspace(-4, 0, 202), np.linspace(4, 0, 202)
    xs, ys = np.meshgrid(xvals, yvals)
    image = hv.Image(np.sin(ys*xs), kdims=['A', 'B'])
    assert Annotator._infer_kdim_dtypes(image) == {'A': np.float64, 'B': np.float64}


def test_infer_kdim_dtype_curve():
    curve = hv.Curve((np.arange('2005-02', '2005-03', dtype='datetime64[D]'), range(28)), kdims=['TIME'])
    assert Annotator._infer_kdim_dtypes(curve) == {'TIME': np.datetime64}


def test_multiplot_add_annotation(multiple_annotators):
    multiple_annotators["annotation1d"].set_range(np.datetime64('2005-02-13'), np.datetime64('2005-02-16'))
    multiple_annotators["annotation2d"].set_range(-0.25, 0.25, -0.1, 0.1)
    multiple_annotators["annotation1d"].add_annotation(description='Multi-plot annotation')
    multiple_annotators["annotation2d"].add_annotation(description='Multi-plot annotation')
    multiple_annotators["annotation1d"].commit()
    multiple_annotators["annotation2d"].commit()


class TestAnnotatorMultipleStringFields:

    def test_insertion_values(self, multiple_fields_annotator):
        start, end = np.datetime64('2022-06-06'), np.datetime64('2022-06-08')
        multiple_fields_annotator.set_range(start, end)
        multiple_fields_annotator.add_annotation(field1='A test field', field2='Another test field')
        commits = multiple_fields_annotator.commit(return_commits=True)
        kwargs = commits[0]['kwargs']
        assert len(commits)==1, 'Only one insertion commit made'
        assert 'uuid' in kwargs.keys(), 'Expected uuid primary key in kwargs'
        kwargs.pop('uuid')
        assert kwargs == dict(field1='A test field', field2='Another test field', start_TIME=start, end_TIME=end)


    def test_commit_insertion(self, multiple_fields_annotator):
        start, end  = np.datetime64('2022-06-06'), np.datetime64('2022-06-08')
        field1 = 'A test field'
        field2 = 'Another test field'
        multiple_fields_annotator.set_range(start, end)
        multiple_fields_annotator.add_annotation(field1=field1, field2=field2)
        multiple_fields_annotator.commit(return_commits=True)

        df = pd.DataFrame({'uuid': pd.Series(multiple_fields_annotator.df.index[0], dtype=object),
                           'start_TIME':[start],
                           'end_TIME':[end],
                           'field1':[field1],
                           'field2':[field2]}
                           ).set_index('uuid')

        sql_df = multiple_fields_annotator.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)


    def test_commit_update(self, multiple_fields_annotator):
        start1, end1  = np.datetime64('2022-06-06'), np.datetime64('2022-06-08')
        start2, end2  = np.datetime64('2023-06-06'), np.datetime64('2023-06-08')
        multiple_fields_annotator.set_range(start1, end1)
        multiple_fields_annotator.add_annotation(field1='Field 1.1', field2='Field 1.2')
        multiple_fields_annotator.set_range(start2, end2)
        multiple_fields_annotator.add_annotation(field1='Field 2.1', field2='Field 2.2')
        multiple_fields_annotator.commit(return_commits=True)
        multiple_fields_annotator.update_annotation_fields(multiple_fields_annotator.df.index[0], field1='NEW Field 1.1')
        multiple_fields_annotator.commit(return_commits=True)
        sql_df = multiple_fields_annotator.connector.load_dataframe()
        assert set(sql_df['field1']) == {'NEW Field 1.1', 'Field 2.1'}


@pytest.mark.parametrize("method", ["new", "same"])
def test_reconnect(method, tmp_path):
    db_path = str(tmp_path / "test.db")

    if method == "new":
        conn1 = SQLiteDB(filename=db_path)
        conn2 = SQLiteDB(filename=db_path)
    elif method == "same":
        conn1 = conn2 = SQLiteDB(filename=db_path)

    # Create annotator with data and commit
    a1 = Annotator(
        spec={"TIME": np.datetime64},
        fields=["description"],
        region_types=["Range"],
        connector=conn1,
    )
    times = pd.date_range("2022-06-09", "2022-06-13")
    for t1, t2 in zip(times[:-1], times[1:]):
        a1.set_range(t1, t2)
        a1.add_annotation(description='A programmatically defined annotation')
    a1.commit(return_commits=True)

    # Save internal dataframes
    a1_df = a1.df.copy()
    a1_region = a1.annotation_table._region_df.copy()
    a1_field = a1.annotation_table._field_df.copy()

    # Add new connector
    a2 = Annotator(
        spec={"TIME": np.datetime64},
        fields=["description"],
        region_types=["Range"],
        connector=conn2,
    )
    a2_df = a2.df.copy()
    a2_region = a2.annotation_table._region_df.copy()
    a2_field = a2.annotation_table._field_df.copy()

    pd.testing.assert_frame_equal(a1_df, a2_df)
    pd.testing.assert_frame_equal(a1_region, a2_region)
    pd.testing.assert_frame_equal(a1_field, a2_field)



from __future__ import annotations

import pandas as pd

from holonote.annotate import AnnotationTable


def test_table_region_df():
    table = AnnotationTable()
    table.load(primary_key_name='id', fields=['test_description'])
    assert len(table._region_df) == 0, 'Should be initialized empty'
    assert tuple(table._region_df.columns) == AnnotationTable.columns

    start = pd.Timestamp('2022-06-17 18:32:48.623476')
    end = pd.Timestamp('2022-06-19 04:44:09.306402')
    regions = [ {'region_type': 'Range',
                    'value': (start, end),
                    'dim1': 'TIME',
                    'dim2': None} ]
    table.add_annotation(regions, id=100, test_description='A test')


    expected = pd.DataFrame([{'region_type':'Range', 'dim1':'TIME', 'dim2': None,
                                'value':(start, end), '_id':100}]).astype({'_id':object})

    pd.testing.assert_frame_equal(table._region_df, expected)



from __future__ import annotations

import holoviews as hv
import numpy as np
import pandas as pd
import param
from bokeh.models.tools import BoxSelectTool, HoverTool
from holoviews.element.selection import Selection1DExpr

from .connector import Connector, SQLiteDB
from .table import AnnotationTable


class Indicator:
    """
    Collection of class methods that express annotation data as final
    displayed (vectorized) HoloViews object.
    """

    range_style = dict(color='red', alpha=0.4, apply_ranges=False)
    point_style = {}
    indicator_highlight = {'alpha':(0.7,0.2)}

    edit_range_style = dict(alpha=0.4, line_alpha=1, line_width=1, line_color='black')
    edit_point_style = {}

    @classmethod
    def indicator_style(cls, range_style, point_style, highlighters):
        return (hv.opts.Rectangles(**dict(range_style, **highlighters)),
                hv.opts.VSpan(**dict(range_style, **highlighters)),
                hv.opts.HSpan(**dict(range_style, **highlighters)))


    @classmethod
    def region_style(cls, edit_range_style, edit_point_style):
        return (hv.opts.Rectangles(**edit_range_style),
                hv.opts.VSpan(**edit_range_style),
                hv.opts.HSpan(**edit_range_style))

    @classmethod
    def points_1d(cls, region_df, field_df, invert_axes=False):
        "Vectorizes point regions to VLines. Note does not support hover info"
        return hv.VLines(list(region_df["value"]))

    @classmethod
    def points_2d(cls, region_df, field_df, invert_axes=False):
        "Vectorizes point regions to VLines * HLines. Note does not support hover info"
        return (hv.VLines([el[0] for el in region_df["value"]])
              * hv.HLines([el[1] for el in region_df["value"]]))

    @classmethod
    def ranges_2d(cls, region_df, field_df, invert_axes=False):
        "Vectorizes an nd-overlay of range_2d rectangles."
        return cls._range_indicators(region_df, field_df, "2d", invert_axes, {})

    @classmethod
    def ranges_1d(cls, region_df, field_df, invert_axes=False, extra_params=None):
        """
        Vectorizes an nd-overlay of range_1d rectangles.

        NOTE: Should use VSpans once available!
        """
        if extra_params is None:
            raise Exception('Extra parameters required until vectorized HSpans/VSpans supported')
        return cls._range_indicators(region_df, field_df, "1d", invert_axes, extra_params)

    @classmethod
    def _range_indicators(cls, region_df, field_df, dimensionality, invert_axes=False, extra_params=None):
        rect_data = []

        mdata_vals = ([None] * len(region_df['_id'])
                      if len(field_df.columns)==0 else field_df.to_dict('records'))
        for id_val, value, mdata in zip(region_df['_id'], region_df["value"], mdata_vals):
            if dimensionality=='1d':
                coords = (value[0], extra_params['rect_min'], value[1], extra_params['rect_max'])
            else:
                coords = (value[0], value[2], value[1], value[3]) # LBRT format

            if None in coords: continue

            mdata_tuple =  () if len(field_df.columns)==0 else tuple(mdata.values())
            rect_data.append(coords + mdata_tuple + (id_val,))

        index_col_name = ['id'] if field_df.index.name is None else [field_df.index.name]
        return hv.Rectangles(rect_data, vdims=list(field_df.columns)+index_col_name) # kdims?




class AnnotatorInterface(param.Parameterized):
    """
    Baseclass that expresses the Python interface of an Annotator
    without using any holoviews components and without requiring
    display. Most of this API centers around how the Annotator interacts
    with the AnnotationTable.
    """

    selected_indices = param.List(default=[], doc="Indices of selected annotations")

    kdim_dtypes = param.Dict(default=None, allow_None=True, doc="""
       Dictionary of one or two key dimension names to dtypes (e.g. int, float, datetime).""")

    connector = param.ClassSelector(class_=Connector, allow_None=False)

    annotation_table = param.ClassSelector(class_=AnnotationTable, allow_None=False)

    region_types = param.ListSelector(default=['Range'], objects=['Range', 'Point'], doc="""
       Enabled region types for the current Annotator.""")

    connector_class = SQLiteDB

    def __init__(self, init=True, **params):
        if "annotation_table" not in params:
            params["annotation_table"] = AnnotationTable()

        connector_kws = {'fields':params.pop('fields')} if 'fields' in params else {}
        connector = params.pop('connector') if 'connector' in params else self.connector_class(**connector_kws)
        super().__init__(connector=connector, **params)
        self._region = {}
        self._last_region = None

        self.annotation_table.register_annotator(self)
        self.annotation_table.add_schema_to_conn(self.connector)

        if init:
            self.load()

    def load(self):
        self.connector._initialize(self.connector.column_schema)
        self.annotation_table.load(self.connector, fields=self.connector.fields)

    @property
    def df(self):
        # NOTE: This code is not general yet and makes assumptions
        # 1) That there is a 1-to-1 mapping after filtering
        field_name = self.connector.primary_key.field_name
        field_df = self.annotation_table._field_df
        fields_columns = list(field_df.columns)
        kdim_names = list(self.kdim_dtypes.keys()) # Needed?? If not, moved to AnnotationTable
        if len(self.kdim_dtypes) == 2:
            dim_mask = self.annotation_table._mask2D(kdim_names)
        elif len(self.kdim_dtypes) == 1:
            dim_mask = self.annotation_table._mask1D(kdim_names)

        region_column_names = []
        views = []
        if 'Range' in self.region_types:
            ranges_df = self.annotation_table._filter(dim_mask, 'Range')
            range_view = field_df.merge(ranges_df, left_index=True, right_on='_id')
            views.append(range_view)
            kdim_names = list(self.kdim_dtypes.keys())
            range_column_names = [f'start[{kdim_names[0]}]', f'end[{kdim_names[0]}]']
            range_view[range_column_names[0]] = range_view['value'].apply(lambda x: x[0])
            range_view[range_column_names[1]] = range_view['value'].apply(lambda x: x[1])
            if len(self.kdim_dtypes) == 2:
                range_column_names.extend([f'start[{kdim_names[1]}]', f'end[{kdim_names[1]}]'])
                range_view[range_column_names[2]] = range_view['value'].apply(lambda x: x[2])
                range_view[range_column_names[3]] = range_view['value'].apply(lambda x: x[3])
            region_column_names.extend(range_column_names)
        if 'Point' in self.region_types:
            points_df = self.annotation_table._filter(dim_mask, 'Point')
            point_view = field_df.merge(points_df, left_index=True, right_on="_id")
            views.append(point_view)
            kdim_names = list(self.kdim_dtypes.keys())
            point_column_names = [f'point[{kdim_names[0]}]']
            point_view[point_column_names[0]] = point_view['value'].apply(lambda x: x[0])
            if len(self.kdim_dtypes) == 2:
                point_column_names.extend([f'point[{kdim_names[1]}]'])
                point_view[point_column_names[1]] = point_view['value'].apply(lambda x: x[1])
            region_column_names.extend(point_column_names)
        view = pd.concat(views)
        view = view.rename(columns={'_id':field_name})
        column_ordering = [field_name, *region_column_names, *fields_columns]
        return view[column_ordering].set_index(field_name)


    def refresh(self, clear=False):
        "Method to update display state of the annotator and optionally clear stale visual state"

    def set_annotation_table(self, annotation_table): # FIXME! Won't work anymore, set_connector??
        self._region = {}
        self.annotation_table = annotation_table
        self.annotation_table.register_annotator(self)
        self.annotation_table._update_index()
        self.snapshot()

    # Selecting annotations

    def select_by_index(self, *inds):
        "Set the selection state by the indices i.e. primary key values"
        self.selected_indices = list(inds)

    @property
    def selected_index(self):
        "Convenience property returning a single selected index (the first one) or None"
        return self.selected_indices[0] if len(self.selected_indices) > 0 else None


    def _get_range_indices_by_position(self, dim1_pos, dim2_pos=None):
        ranges = self.annotation_table._region_df[self.annotation_table._region_df['region_type']=='Range']
        mask = ranges['value'].apply(lambda tpl: tpl[0] <= dim1_pos < tpl[1])
        if dim2_pos is not None:
            mask_dim2 = ranges['value'].apply(lambda tpl: tpl[2] <= dim2_pos < tpl[3])
            mask = mask & mask_dim2
        if not len(mask):
            return []
        return list(ranges[mask]['_id'])


    def get_indices_by_position(self, dim1_pos, dim2_pos=None):
        "Return primary key values matching given position in data space"
        # Lots TODO! 2 Dimensions, different annotation types etc.
        range_matches = self._get_range_indices_by_position(dim1_pos, dim2_pos)
        event_matches = [] # TODO: Needs hit testing or similar for point events
        return range_matches + event_matches

    @property
    def region(self):
        return self._region

    def set_range(self, startx, endx, starty=None, endy=None):
        if len(self.kdim_dtypes) == 2 and None in [starty, endy ]:
            raise ValueError('Two key dimensions specified: both starty and endy arguments required')
        if 'Range' not in self.region_types:
            raise ValueError(f'Range region types not enabled as region_types={self.region_types}')
        if None in [starty, endy] and ([starty, endy] != [None, None]):
            raise ValueError('Both starty and endy need to be non-None')

        value = (startx, endx) if(None in [starty, endy]) else (startx, endx, starty, endy)
        kdims = list(self.kdim_dtypes.keys())
        self._set_region('Range', value, *kdims)

    def set_point(self, posx, posy=None):
        if 'Point' not in self.region_types:
            raise ValueError(f'Point region types not enabled as region_types={self.region_types}')

        value = (posx,None) if posy is None else (posx, posy)
        kdims = list(self.kdim_dtypes.keys())
        self._set_region('Point', value,  *kdims)


    def _set_region(self, region_type, value=None, dim1=None, dim2=None):
        "Use _set_region(None) to clear currently defined region"
        if (region_type, value, dim1, dim2) == (None, None, None, None):
            self._region = {}
        elif None not in [value, dim1]:
            self._region = {'region_type':region_type, 'value':value, 'dim1':dim1, 'dim2':dim2}
        else:
            raise Exception('Both value and dim1 required for non-None region type')

    def add_annotation(self, **fields):  #   Rename box to range.
        "Primary key specification is optional"
        if self.connector.primary_key.field_name not in fields:
            index_val = self.connector.primary_key(self.connector,
                                                   list(self.annotation_table._field_df.index))
            fields[self.connector.primary_key.field_name] = index_val

        if self.region != self._last_region:
            if len(self.annotation_table._annotators)>1:
                raise AssertionError('Multiple annotation instances attached to the connector: '
                                     'Call add_annotation directly from the associated connector.')
            self.annotation_table.add_annotation('annotator-regions', **fields)
        self._last_region = self._region

    def update_annotation_region(self, index):
        self.annotation_table.update_annotation_region(index)


    def update_annotation_fields(self, index, **fields):
        self.annotation_table.update_annotation_fields(index, **fields)


    def delete_annotation(self, index):
        self.annotation_table.delete_annotation(index)

    # Defining initial annotations

    def define_fields(self, fields_df, preserve_index=False):
        if not preserve_index:
            indices = [self.connector.primary_key(self.connector) for el in range(len(fields_df))]
            index_mapping = {old:new for old, new in zip(fields_df.index, indices)}
            fields_df = fields_df.set_index(pd.Series(indices,
                                                      name=self.connector.primary_key.field_name))
        else:
            index_mapping = {ind:ind for ind in fields_df.index}
        self.annotation_table.define_fields(fields_df, index_mapping)

    def define_ranges(self, startx, endx, starty=None, endy=None, dims=None):
        if 'Range' not in self.region_types:
            raise ValueError('Range region types not enabled')
        if dims is None:
            raise ValueError('Please specify dimension annotated by defined ranges')

        self.annotation_table.define_ranges(dims, startx, endx, starty, endy)

    def define_points(self, posx, posy=None, dims=None):
        if 'Point' not in self.region_types:
            raise ValueError('Point region types not enabled')

        if dims is None:
            raise ValueError('Please specify dimension annotated by defined ranges')
        self.annotation_table.define_points(dims, posx, posy=posy)


    # Snapshotting and reverting

    def revert_to_snapshot(self):
        self.annotation_table.revert_to_snapshot()

    def snapshot(self):
        self.annotation_table.snapshot()

    def commit(self, return_commits=False):
        # self.annotation_table.initialize_table(self.connector)  # Only if not in params
        commits = self.annotation_table.commits(self.connector)
        if return_commits:
            return commits


class AnnotatorPlot(AnnotatorInterface):
    rect_min = param.Number(default=-1000, doc="Temporary parameter until vectorized element fully supported")

    rect_max = param.Number(default=1050, doc="Temporary parameter until vectorized element fully supported")

    indicator = Indicator

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._selection_enabled = True
        self._editable_enabled = True
        self._selected_values = []
        self._selected_options = []

        transient=False
        self._edit_streams = [
                hv.streams.BoundsXY(transient=transient),
                hv.streams.SingleTap(transient=transient),
                hv.streams.Lasso(transient=transient),
        ]

    @property
    def edit_tools(self):
        tools = []
        if 'Range' in self.region_types and len(self.kdim_dtypes)==1:
            tools.append(BoxSelectTool(dimensions="width"))
        elif 'Range' in self.region_types and len(self.kdim_dtypes)==2:
            tools.append(BoxSelectTool())

        if 'Point' in self.region_types:
            tools.append('tap')

        return tools

    @classmethod
    def _infer_kdim_dtypes(cls, element):
        if not isinstance(element, hv.Element):
            raise ValueError('Supplied object {element} is not a bare HoloViews Element')
        kdim_dtypes = {}
        for kdim in element.dimensions(selection='key'):
            kdim_dtypes[str(kdim)] = type(element.dimension_values(kdim)[0])
        return kdim_dtypes

    def clear_indicated_region(self):
        "Clear any region currently indicated on the plot by the editor"
        self._edit_streams[0].event(bounds=None)
        self._edit_streams[1].event(x=None, y=None)
        self._edit_streams[2].event(geometry=None)
        self._set_region(None)

    @property
    def selection_element(self):
        if self.element is None:
            kdims = list(self.kdim_dtypes.keys())
            kdim_dtype = next(iter(self.kdim_dtypes.values()))
            return hv.Curve(([kdim_dtype(), kdim_dtype()],
                             [0,1]), kdims=kdims) # Note: Any concrete Selection1dExpr will do...

        if isinstance(self.element, (hv.NdOverlay, hv.Overlay)):
            expr_1d_set = set(isinstance(el, Selection1DExpr) for el in self.element)
            if len(expr_1d_set) == 1:
                return next(iter(self.element))
            else:
                raise ValueError("Inconsistent selection types in overlay")
        else:
            return self.element

    @property
    def selection_enabled(self):
        return self._selection_enabled

    @selection_enabled.setter
    def selection_enabled(self, enabled):
        self._selection_enabled = enabled
        if not enabled:
            self.select_by_index()

    @property
    def editable_enabled(self):
        return self._editable_enabled

    @editable_enabled.setter
    def editable_enabled(self, enabled):
        self._editable_enabled = enabled
        if not enabled:
            self.clear_indicated_region()


    def _filter_stream_values(self, bounds, x, y, geometry):
        if not self._editable_enabled:
            return (None, None, None, None)
        if 'Range' not in self.region_types:
            bounds = None

        # If selection enabled, tap stream used for selection not for creating point regions
        if 'Point' in self.region_types and self.selection_enabled:
            x, y = None, None
        elif 'Point' not in self.region_types:
            x, y = None, None

        return bounds, x, y, geometry


    def _make_selection_dmap(self):
        def inner(bounds, x, y, geometry):
            bounds, x, y, geometry = self._filter_stream_values(bounds, x, y, geometry)

            info = self.selection_element._get_selection_expr_for_stream_value(
                bounds=bounds, x=x, y=y, geometry=geometry
            )
            (dim_expr, bbox, region_element) = info

            self._selection_info = {
                "dim_expr": dim_expr,
                "bbox": bbox,
                "x": x,
                "y": y,
                "geometry": geometry,
                "region_element": region_element,
            }

            if bbox is not None:
                dim1, dim2 = None, None
                dim_names = list(bbox.keys())
                dim1 = dim_names[0]
                value = bbox[dim1]

                if len(bbox)== 2:
                    dim2 = dim_names[1]
                    start1,end1 = value
                    start2,end2 = bbox[dim2]
                    value = (start1, end1, start2,end2)

                self._set_region('Range', value=value, dim1=dim1, dim2=dim2)

            if None not in [x,y]:
                kdims = list(self.kdim_dtypes.keys())
                value = (x,None) if len(self.kdim_dtypes) == 1 else (x,y)
                self._set_region('Point', value,  *kdims)

            return region_element

        return hv.DynamicMap(inner, streams=self._edit_streams)

    def register_tap_selector(self, element):
        def tap_selector(x,y): # Tap tool must be enabled on the element
            # Only select the first

            if len(self.kdim_dtypes) == 1:
                indices = self.get_indices_by_position(x)
            elif len(self.kdim_dtypes) == 2:
                indices = self.get_indices_by_position(x, y)
            if indices:
                self.select_by_index(indices[0])
            else:
                self.select_by_index()
        tap_stream = hv.streams.Tap(source=element, transient=True)
        tap_stream.add_subscriber(tap_selector)
        return element


    def indicators(self):

        if self.element is not None:
            self.register_tap_selector(self.element)

        def inner(_count):
            return self.static_indicators

        return hv.DynamicMap(inner, streams=[self._annotation_count_stream])

    def overlay(self,
                element = True,
                indicators=True,
                editor=True,
                range_style=None,
                point_style=None,
                edit_range_style=None,
                edit_point_style=None,
                highlight=None):

        if range_style is None: range_style = Indicator.range_style
        if point_style is None: point_style = Indicator.point_style
        if edit_range_style is None: edit_range_style = Indicator.edit_range_style
        if edit_point_style is None: edit_point_style = Indicator.edit_point_style
        if highlight is None: highlight=Indicator.indicator_highlight

        highlighters = {opt:self.selected_dim_expr(v[0], v[1]) for opt,v in highlight.items()}
        indicator_style = Indicator.indicator_style(range_style, point_style, highlighters)
        region_style = Indicator.region_style(edit_range_style, edit_point_style)

        layers = []
        if element is True and self.element is None:
            raise ValueError('Annotator was not supplied an element to display.')
        elif element not in [True, False]:
            element = self.register_tap_selector(element)
            layers.append(element.opts(tools=self.edit_tools, active_tools=["box_select"]))
        else:
            layers.append(self.element.opts(tools=self.edit_tools))

        if indicators:
            layers.append(self.indicators().opts(*indicator_style))
        if editor:
            layers.append(self.region_editor().opts(*region_style))
        return hv.Overlay(layers).collate()

    def region_editor(self):
        return self._make_selection_dmap()

    def _build_hover_tool(self):
        # FIXME: Not generalized yet - assuming range
        extra_cols = [(col, '@{%s}' % col.replace(' ','_')) for col in self.annotation_table._field_df.columns]
        region_tooltips = []
        region_formatters = {}
        for direction, kdim in zip(['x','y'], self.kdim_dtypes.keys()):
            if self.kdim_dtypes[kdim] is np.datetime64:
                region_tooltips.append((f'start {kdim}', f'@{direction}0{{%F}}'))
                region_tooltips.append((f'end {kdim}', f'@{direction}1{{%F}}'))
                region_formatters[f'@{direction}0'] = 'datetime'
                region_formatters[f'@{direction}1'] = 'datetime'
            else:
                region_tooltips.append((f'start {kdim}', f'@{direction}0'))
                region_tooltips.append((f'end {kdim}', f'@{direction}1'))

        return HoverTool(tooltips=region_tooltips+extra_cols,
                         formatters = region_formatters)


    def _point_indicators(self, filtered_df, dimensionality, invert_axes=False):
        if dimensionality == '1d':
            return Indicator.points_1d(filtered_df, None, invert_axes=invert_axes)
        else:
            return Indicator.points_2d(filtered_df, None, invert_axes=invert_axes)

    def _range_indicators(self, filtered_df, dimensionality, invert_axes=False):
        fields_mask = self.annotation_table._field_df.index.isin(filtered_df['_id'])
        field_df = self.annotation_table._field_df[fields_mask]  # Currently assuming 1-to-1
        if dimensionality == '1d':
            extra_params = {'rect_min':self.rect_min, 'rect_max':self.rect_max} # TODO: Remove extra_params!
            vectorized = Indicator.ranges_1d(filtered_df, field_df, invert_axes=invert_axes,
                                             extra_params=extra_params)
        else:
            vectorized = Indicator.ranges_2d(filtered_df, field_df, invert_axes=invert_axes)
        return vectorized.opts(tools=[self._build_hover_tool()])


    @property
    def static_indicators(self):
        invert_axes = False  # Not yet handled
        kdims = list(self.kdim_dtypes.keys())
        if len(kdims) == 1:
            dim_mask = self.annotation_table._mask1D(kdims)
            points_df = self.annotation_table._filter(dim_mask, "Point")
            ranges_df = self.annotation_table._filter(dim_mask, "Range")
            if len(points_df) == 0:
                return self._range_indicators(ranges_df, '1d', invert_axes=invert_axes)
            elif len(ranges_df) == 0:
                return self._point_indicators(points_df, '1d', invert_axes=invert_axes)
            else:
                raise NotImplementedError  # FIXME: Both in overlay

        if len(kdims) > 1:
            dim_mask = self.annotation_table._mask2D(kdims)
            points_df = self.annotation_table._filter(dim_mask, "Point")
            ranges_df = self.annotation_table._filter(dim_mask, "Range")
            if len(points_df) == 0:
                return self._range_indicators(ranges_df, '2d', invert_axes=invert_axes)
            elif len(ranges_df) == 0:
                return self._point_indicators(points_df, '2d', invert_axes=invert_axes)
            else:
                raise NotImplementedError  # FIXME: Both in overlay


    def set_range(self, startx, endx, starty=None, endy=None):
        super().set_range(startx, endx, starty, endy)
        self.show_region()

    def set_point(self, posx, posy=None):
        super().set_point(posx, posy)
        self.show_region()


class Annotator(AnnotatorPlot):
    """
    An annotator displays the contents of an AnnotationTable and
    provides the means to manipulate view the corresponding contents,
    add new annotations and update existing annotations.
    """


    _count = param.Integer(default=0, precedence=-1)


    def __init__(self, spec, **params):
        """
        The spec argument must be an element or a dictionary of kdim dtypes
        """
        self.element = None
        if not isinstance(spec, dict):
            self.element = spec.clone() # Clone so we can add a tap tool and trust our copy.

        super().__init__(
            kdim_dtypes=spec if self.element is None else self._infer_kdim_dtypes(spec),
            **params,
        )
        self._selection_info = {}

        self._annotation_count_stream = hv.streams.Params(
            parameterized=self,
            parameters=['_count'],
            transient=True,
        )

    def refresh(self, clear=False) -> None:
        hv.streams.Stream.trigger([self._annotation_count_stream])
        if clear:
            self.clear_indicated_region()

    def set_annotation_table(self, annotation_table):
        self.select_by_index()
        self._set_region(None)
        super().set_annotation_table(annotation_table)
        self.refresh(clear=True)

    def selected_dim_expr(self, selected_value, non_selected_value):
        self._selected_values.append(selected_value)
        self._selected_options.append({i:selected_value for i in self.selected_indices})
        index_name = ('id' if (self.annotation_table._field_df.index.name is None)
                      else self.annotation_table._field_df.index.name)
        return hv.dim(index_name).categorize(
            self._selected_options[-1], default=non_selected_value)

    def clear_edits(self):
        super().clear_edits()
        self.clear_indicated_region()

    @property
    def dim_expr(self):
        return self._selection_info["dim_expr"]

    def show_region(self):
        if self.region != {}:
            if len(self.region['value']) == 2:
                bounds = (self.region['value'][0], 0,
                      self.region['value'][1], 1)
            else:
                bounds = (self.region['value'][0], self.region['value'][2],
                          self.region['value'][1], self.region['value'][3])
            self._edit_streams[0].event(bounds = bounds)

    # Methods overriding the baseclass to add extra interactivity

    def select_by_index(self, *inds):
        "Set the selection state by the indices i.e. primary key values"

        if not self.selection_enabled:
            inds = ()
        super().select_by_index(*inds)

        for d, val in zip(self._selected_options, self._selected_values):
            d.clear()
            for ind in inds:
                d[ind] = val

        self.refresh()

    def add_annotation(self, **fields):
        super().add_annotation(**fields)
        self.refresh(clear=True)

    def update_annotation_region(self, index):
        super().update_annotation_region(index)
        self.refresh()


    def update_annotation_fields(self, index, **fields):
        super().update_annotation_fields(index, **fields)
        self.refresh()


    def delete_annotation(self, *indices):
        if len(indices)==0:
            raise ValueError('At least one index must be specified to delete annotations')
        for index in indices:
            super().delete_annotation(index)
        self.refresh()

    def define_fields(self, fields_df, preserve_index=False):
        """
        If insert_index is True, the index values are inserted as primary key values
        """
        super().define_fields(fields_df, preserve_index=preserve_index)

    # Pass-through methods that get dimensionality from self.element
    def define_ranges(self, startx, endx, starty=None, endy=None, dims=None):
        "Define ranges using element kdims as default dimensions."
        if dims is None:
            dims = list(self.kdim_dtypes.keys())
        super().define_ranges(startx, endx, starty=starty, endy=endy, dims=dims)
        self.refresh()


    def define_points(self, posx, posy=None, dims=None):
        "Define points using element kdims as default dimensions."
        if dims is None:
            dims = list(self.kdim_dtypes.keys())
        super().define_points(posx, posy=posy, dims=dims)
        self.refresh()

    @property
    def has_snapshot(self):
        return self.annotation_table.has_snapshot

    def revert_to_snapshot(self):
        super().revert_to_snapshot()
        self.refresh()
